from dovado_rtl.antlr.hdl_representation import (
    Entity,
    ParameterTypeEnum,
    PortDirectionEnum,
    PortTypeEnum,
    TopLevel,
    Port,
    Parameter,
    PortDirection,
    PortType,
    ParameterType,
)
from dovado_rtl.antlr.generated.vhdl.vhdlParserVisitor import vhdlParserVisitor
from dovado_rtl.antlr.generated.vhdl.vhdlParser import vhdlParser
from typing import List, Iterator, Optional, Union, Tuple
from itertools import repeat


class VhdlEntityVisitor(vhdlParserVisitor):
    def __init__(self):
        self.top_level: TopLevel = TopLevel([], [])
        self.entities: List[Entity] = []

    def visitDesign_file(self, ctx: vhdlParser.Design_fileContext):
        for d in ctx.design_unit():
            self.visitDesign_unit(d)

    def visitDesign_unit(self, ctx: vhdlParser.Design_unitContext):
        self.visitContext_clause(ctx.context_clause())
        self.visitLibrary_unit(ctx.library_unit())

    def visitContext_clause(self, ctx: vhdlParser.Context_clauseContext):
        for i in ctx.context_item():
            self.visitContext_item(i)

    def visitContext_item(self, ctx: vhdlParser.Context_itemContext):
        if ctx.library_clause():
            self.visitLibrary_clause(ctx.library_clause())
        if ctx.use_clause():
            self.visitUse_clause(ctx.use_clause())
        if ctx.context_reference():
            pass

    def visitLibrary_clause(self, ctx: vhdlParser.Library_clauseContext):
        for ln in self.visitLogical_name_list(ctx.logical_name_list()):
            self.top_level.add_library(ln)

    def visitUse_clause(self, ctx: vhdlParser.Use_clauseContext):
        for n in ctx.selected_name():
            self.top_level.add_used(self.visitSelected_name(n))

    def visitLogical_name_list(
        self, ctx: vhdlParser.Logical_name_listContext
    ) -> Iterator[str]:
        return self.visitIdentifier_list(ctx.identifier_list())

    def visitIdentifier_list(
        self, ctx: vhdlParser.Identifier_listContext
    ) -> Iterator[str]:
        for i in ctx.identifier():
            yield self.visitIdentifier(i)

    def visitSelected_name(self, ctx: vhdlParser.Selected_nameContext) -> str:
        out = self.visitIdentifier(ctx.identifier())
        for s in ctx.suffix():
            suff = self.visitSuffix(s)
            if suff:
                out += "." + suff
        return out

    def visitSuffix(self, ctx: vhdlParser.SuffixContext) -> Optional[str]:
        if ctx.name_literal():
            return self.visitName_literal(ctx.name_literal())
        if ctx.KW_ALL():
            return ctx.KW_ALL().getText()

    def visitIdentifier(self, ctx: vhdlParser.IdentifierContext):
        return (
            ctx.BASIC_IDENTIFIER().getText()
            if ctx.BASIC_IDENTIFIER()
            else ctx.EXTENDED_IDENTIFIER()
        )

    def visitLibrary_unit(self, ctx: vhdlParser.Library_unitContext):
        if ctx.primary_unit():
            self.visitPrimary_unit(ctx.primary_unit())
        if ctx.secondary_unit():
            # architecture body and package body ignored
            pass

    def visitPrimary_unit(self, ctx: vhdlParser.Primary_unitContext):
        if ctx.entity_declaration():
            self.visitEntity_declaration(ctx.entity_declaration())
        else:
            pass

    def visitEntity_declaration(
        self, ctx: vhdlParser.Entity_declarationContext
    ):
        self.entities.append(Entity(self.visitIdentifier(ctx.identifier()[0])))
        if ctx.generic_clause():
            self.visitGeneric_clause(ctx.generic_clause())
        if ctx.port_clause():
            self.visitPort_clause(ctx.port_clause())

    def visitGeneric_clause(self, ctx: vhdlParser.Generic_clauseContext):
        self.visitGeneric_list(ctx.generic_list())

    def visitPort_clause(self, ctx: vhdlParser.Port_clauseContext):
        self.visitPort_list(ctx.port_list())

    def visitGeneric_list(self, ctx: vhdlParser.Generic_listContext):
        for i in self.visitInterface_list(ctx.interface_list()):
            p_type = i[1]
            if isinstance(p_type, ParameterType):
                self.entities[-1].add_parameter(Parameter(i[0], p_type))

    def visitPort_list(self, ctx: vhdlParser.Port_listContext):
        for i in self.visitInterface_list(ctx.interface_list()):
            p_dir = i[1]
            p_type = i[2]
            if (
                len(i) == 3
                and isinstance(p_dir, PortDirection)
                and isinstance(p_type, PortType)
            ):
                self.entities[-1].add_port(Port(i[0], p_dir, p_type))
            else:
                pass

    def visitInterface_list(
        self, ctx: vhdlParser.Interface_listContext
    ) -> Union[
        List[Tuple[str, ParameterType]],
        List[Tuple[str, PortDirection, PortType]],
    ]:
        out = []
        for i in ctx.interface_element():
            for e in self.visitInterface_element(i):
                out.append(e)
        return out

    def visitInterface_element(
        self, ctx: vhdlParser.Interface_elementContext
    ) -> Union[
        Union[
            Iterator[Tuple[str, ParameterType]],
            Union[
                List[Tuple[str, ParameterType]],
                List[Tuple[str, PortDirection, PortType]],
            ],
        ],
        None,
    ]:
        return self.visitInterface_declaration(ctx.interface_declaration())

    def visitInterface_declaration(
        self, ctx: vhdlParser.Interface_declarationContext
    ) -> Union[
        Union[
            Iterator[Tuple[str, ParameterType]],
            Union[
                List[Tuple[str, ParameterType]],
                List[Tuple[str, PortDirection, PortType]],
            ],
        ],
        None,
    ]:
        if ctx.interface_object_declaration():
            return self.visitInterface_object_declaration(
                ctx.interface_object_declaration()
            )
        else:
            return []

    def visitInterface_object_declaration(
        self, ctx: vhdlParser.Interface_object_declarationContext
    ) -> Union[
        Union[
            Iterator[Tuple[str, ParameterType]],
            Union[
                List[Tuple[str, ParameterType]],
                List[Tuple[str, PortDirection, PortType]],
            ],
        ],
        None,
    ]:
        if ctx.interface_constant_declaration():
            return self.visitInterface_constant_declaration(
                ctx.interface_constant_declaration()
            )
        if ctx.interface_signal_declaration():
            return self.visitInterface_signal_declaration(
                ctx.interface_signal_declaration()
            )

    def visitInterface_constant_declaration(
        self, ctx: vhdlParser.Interface_constant_declarationContext
    ) -> Iterator[Tuple[str, ParameterType]]:
        for i, t in zip(
            self.visitIdentifier_list(ctx.identifier_list()),
            repeat(self.visitSubtype_indication(ctx.subtype_indication())),
        ):
            if isinstance(t, ParameterType):
                yield i, t
            else:
                pass

    def visitInterface_signal_declaration(
        self, ctx: vhdlParser.Interface_signal_declarationContext
    ) -> Union[
        List[Tuple[str, ParameterType]],
        List[Tuple[str, PortDirection, PortType]],
    ]:
        out = []
        if not ctx.signal_mode():
            for i, t in zip(
                self.visitIdentifier_list(ctx.identifier_list()),
                repeat(self.visitSubtype_indication(ctx.subtype_indication())),
            ):
                if isinstance(t, ParameterType):
                    out.append((i, t))
                else:
                    pass
            return out
        else:
            for i, s, t in zip(
                self.visitIdentifier_list(ctx.identifier_list()),
                repeat(self.visitSignal_mode(ctx.signal_mode())),
                repeat(self.visitSubtype_indication(ctx.subtype_indication())),
            ):
                if isinstance(t, PortType):
                    out.append((i, s, t))
                else:
                    pass
            return out

    def visitSignal_mode(
        self, ctx: vhdlParser.Signal_modeContext
    ) -> PortDirection:
        return (
            PortDirection(PortDirectionEnum.INPUT, "input")
            if ctx.KW_IN()
            else PortDirection(PortDirectionEnum.OUTPUT, "output")
            if ctx.KW_OUT()
            else PortDirection(PortDirectionEnum.OTHER, "buffer")
            if ctx.KW_BUFFER()
            else PortDirection(PortDirectionEnum.OTHER, "inout")
            if ctx.KW_INOUT()
            else PortDirection(PortDirectionEnum.OTHER, "linkage")
        )

    def visitSubtype_indication(
        self, ctx: vhdlParser.Subtype_indicationContext
    ) -> Union[PortType, ParameterType, None]:
        type = self.visitType_mark(ctx.type_mark())
        if type == "std_logic":
            return PortType(PortTypeEnum.SCALAR, "std_logic")
        if type == "std_ulogic":
            return PortType(PortTypeEnum.SCALAR, "std_ulogic")
        if type == "std_ulogic_vector":
            return PortType(PortTypeEnum.VECTOR, "std_ulogic_vector")
        elif type == "std_logic_vector":
            return PortType(PortTypeEnum.VECTOR, "std_logic_vector")
        elif type == "integer":
            return ParameterType(ParameterTypeEnum.INTEGER, "integer")
        elif type == "natural":
            return ParameterType(ParameterTypeEnum.INTEGER, "natural")
        elif type == "positive":
            return ParameterType(ParameterTypeEnum.INTEGER, "positive")
        elif type == "signed":
            return ParameterType(ParameterTypeEnum.INTEGER, "signed")
        elif type == "unsigned":
            return ParameterType(ParameterTypeEnum.INTEGER, "unsigned")

        else:
            pass

    def visitType_mark(
        self, ctx: vhdlParser.Type_markContext
    ) -> Optional[str]:
        return self.visitName(ctx.name())

    def visitName(self, ctx: vhdlParser.NameContext) -> Optional[str]:
        if ctx.name_literal():
            return self.visitName_literal(ctx.name_literal())
        if ctx.external_name():
            return self.visitExternal_name(ctx.external_name())
        else:
            return self.visitName(ctx.name())
            # skipping range in e.g. std_logic_vector

    def visitName_literal(
        self, ctx: vhdlParser.Name_literalContext
    ) -> Optional[str]:
        if ctx.identifier():
            return self.visitIdentifier(ctx.identifier())
        if ctx.operator_symbol():
            return self.visitOperator_symbol(ctx.operator_symbol())
        if ctx.CHARACTER_LITERAL():
            return ctx.CHARACTER_LITERAL().getText()

    def visitOperator_symbol(
        self, ctx: vhdlParser.Operator_symbolContext
    ) -> str:
        return ctx.STRING_LITERAL().getText()

    def visitExternal_name(self, ctx: vhdlParser.External_nameContext) -> str:
        return (
            ctx.SHIFT_LEFT().getText() + ctx.KW_VARIABLE().getText()
            if ctx.KW_VARIABLE()
            else ctx.KW_CONSTANT().getText()
            if ctx.KW_CONSTANT
            else ctx.KW_SIGNAL().getText()
            + self.visitExternal_pathname(ctx.external_pathname())
            + ctx.COLON().getText()
            + self.visitSubtype_indication(ctx.subtype_indication())
            + ctx.SHIFT_RIGHT().getText()
        )

    def visitExternal_pathname(
        self, ctx: vhdlParser.External_pathnameContext
    ) -> Optional[str]:
        if ctx.package_pathname():
            return self.visitPackage_pathname(ctx.package_pathname())
        if ctx.absolute_pathname():
            return self.visitAbsolute_pathname(ctx.absolute_pathname())
        if ctx.relative_pathname():
            return self.visitRelative_pathname(ctx.relative_pathname())

    def visitPackage_pathname(
        self, ctx: vhdlParser.Package_pathnameContext
    ) -> str:
        out = ctx.AT().getText()
        for i in ctx.identifier()[:-1]:
            out += self.visitIdentifier(i) + "."
        return out + self.visitIdentifier(ctx.identifier()[-1])

    def visitAbsolute_pathname(
        self, ctx: vhdlParser.Absolute_pathnameContext
    ) -> str:
        return "." + self.visitPartial_pathname(ctx.partial_pathname())

    def visitRelative_pathname(
        self, ctx: vhdlParser.Relative_pathnameContext
    ) -> str:
        return (
            (ctx.UP().getText() + ".")
            if ctx.UP()
            else "" + self.visitPartial_pathname(ctx.partial_pathname())
        )

    def visitPartial_pathname(
        self, ctx: vhdlParser.Partial_pathnameContext
    ) -> str:
        out = ""
        for p in ctx.pathname_element():
            out += self.visitPathname_element(p) + "."
        return out + self.visitIdentifier(ctx.identifier())

    def visitPathname_element(
        self, ctx: vhdlParser.Pathname_elementContext
    ) -> str:
        if ctx.expression():
            print("Expression in pathname are not supported, skipping")
        return self.visitLabel(ctx.label())

    def visitLabel(self, ctx: vhdlParser.LabelContext) -> str:
        return self.visitIdentifier(ctx.identifier())



from click import BadParameter
import typer
from typing import Tuple, List
from pathlib import Path
from dovado_rtl.doc_parsing import get_directives, get_directives_paragraph
from dovado_rtl.src_parsing import SourceParser
from dovado_rtl.vivado_interaction import get_parts
from dovado_rtl.antlr.hdl_representation import PortDirectionEnum
from dovado_rtl.enums import RegressionModel


def validate_nearest_distance(value: int) -> int:
    if value < 0:
        raise typer.BadParameter(
            "Invalid distance value: "
            + str(value)
            + " non-negative integer expected"
        )
    return value


def validate_estimator(value: str) -> str:
    if value.upper() not in RegressionModel.__members__.keys():
        raise typer.BadParameter(
            "Invalid estimator "
            + value
            + " available estimators are "
            + str(set(RegressionModel.__members__.keys()))
        )
    return value


def validate_board(value: str) -> str:
    parts = get_parts()
    if value in set(parts.keys()):
        return value
    raise typer.BadParameter(
        "Invalid board " + value + " available boards are: " + str(parts)
    )


def validate_parameters(ctx: typer.Context, value: List[str]) -> List[str]:

    file_path = Path(ctx.params["file_path"]).absolute()
    if not ctx.obj:
        parsed_src = SourceParser(
            str(SourceParser.get_project_path(file_path)), str(file_path)
        )
        parsed_src.set_entity(str(parsed_src.get_entities()[0].get_name()))
        ctx.obj = parsed_src
    else:
        parsed_src: SourceParser = ctx.obj
    parameter_name_list = [p.get_name() for p in parsed_src.get_parameters()]
    for chosen_parameter in value:
        if chosen_parameter not in parameter_name_list:
            raise typer.BadParameter(
                "Chosen parameter: "
                + chosen_parameter
                + " is not among available parameters "
                + str(parameter_name_list)
            )
    return value


def validate_clock_port(ctx: typer.Context, value: str) -> str:
    parsed_src: SourceParser = ctx.obj
    for port in parsed_src.get_ports():
        if (
            port.get_name() == value
            and port.get_direction().direction is PortDirectionEnum.INPUT
        ):
            return value
    raise typer.BadParameter(
        "Chosen clock port: "
        + value
        + " is not among available input ports "
        + str(
            [
                p.get_name()
                for p in parsed_src.get_ports()
                if p.get_direction().direction is PortDirectionEnum.INPUT
            ]
        )
    )


def validate_directives(value: Tuple[str, str, str]):
    commands = ["synthesis", "place", "route"]
    for c in commands:
        directive = value[commands.index(c)]
        directives = get_directives((get_directives_paragraph(c)))
        if directive not in directives:
            raise typer.BadParameter(
                "Chosen directive for "
                + c
                + ": "
                + directive
                + " is not among available directives for "
                + c
                + " "
                + str(directives)
            )
    return value


def validate_target_clock(value: float) -> float:
    if value > 0:
        return value
    raise BadParameter(
        "Target clock must be positive, invalid target clock input: "
        + str(value)
    )


def validate_param_ranges(value: List[int]) -> List[int]:
    if len(value) % 2 == 0:
        return value
    raise typer.BadParameter(
        "Parameter ranges list must be of even length, "
        + "each odd element is the start of the range and the next one is the end of the range"
    )



import re
from pathvalidate import is_valid_filepath
from os import path
from pathlib import Path
from collections import OrderedDict
from typing import Tuple, Optional, Set, List, Dict
import dovado_rtl.vivado_interaction as vivado
import dovado_rtl.doc_parsing as doc
from dovado_rtl.config import Configuration
from dovado_rtl.simple_types import IsIncremental, Metric
from dovado_rtl.enums import RTL, StopStep, ImplementationStep
from dovado_rtl.src_parsing import SourceParser
from dovado_rtl.antlr.hdl_representation import (
    Entity,
    Parameter,
    Port,
    PortDirection,
    PortDirectionEnum,
    PortType,
    PortTypeEnum,
)


def list_rtl_files(src_path: Path) -> Set[Path]:
    rtl_extensions = ["vhd", "vhdl", "v", "sv"]
    files = []
    for extension in rtl_extensions:
        files += [
            pth for pth in src_path.rglob("*." + extension) if pth.is_file
        ]
    return set(files)


def ask_code_dir() -> str:
    while True:
        default = "./"
        user_input = input(
            "Enter path to code folder" + " [default = " + "./" + "]: "
        )
        if (
            is_valid_filepath(user_input)
            and path.exists(user_input)
            and path.isdir(user_input)
            and (list_rtl_files(Path(user_input)))
        ) or user_input == "":
            return default if user_input == "" else user_input

        print("Invalid path, ensure there are no typos\n")


def get_top_module(
    src_folder: str, module_file: str, config: Configuration
) -> Tuple[Entity, str, SourceParser]:
    for src in list_rtl_files(Path(src_folder)):
        if src.name == module_file:
            sp = SourceParser(src_folder, str(src))
            entities = sp.get_entities()
            if len(entities) > 1:
                print(
                    "Discarding all entities except the first from "
                    + str(src)
                    + " please refactor your code for a more consistent behaviour."
                    + " Put your target module in a file by itself"
                )
            return entities[0], str(src), sp
    raise ValueError("Could not find " + module_file + " in " + src_folder)


def ask_top_module(
    src_folder: str, config: Configuration
) -> Tuple[Entity, str, SourceParser]:
    while True:
        user_input = input("Enter top file (including suffix): ")
        try:
            return get_top_module(src_folder, user_input, config,)
        except ValueError as e:
            print(e)
            continue


def ask_part() -> str:
    parts = vivado.get_parts()
    default = list(parts.keys())[0]
    while True:
        user_input = input(
            "Enter part (input ? to have a list of available parts)"
            + "[default = "
            + default
            + "]: "
        )
        if user_input in parts or user_input == "":
            return default if user_input == "" else user_input
        if user_input == "?":
            print("Available parts\n" + str(parts))
            continue
        print("Part not available.")


def ask_stop_step() -> StopStep:
    default = "synthesis"
    while True:
        user_input = input(
            "Would you like to get utilization and frequency after logical"
            + " synthesis or after implementation?"
            + "(allowed answers = synthesis/implementation)"
            + "[default = "
            + default
            + "]: "
        )
        if (
            user_input == "synthesis"
            or user_input == "implementation"
            or user_input == ""
        ):
            stop_step = default if user_input == "" else user_input
            return (
                StopStep.SYNTHESIS
                if stop_step == "synthesis"
                else StopStep.IMPLEMENTATION
            )


def ask_incremental_mode(stop_step: StopStep) -> IsIncremental:
    default = "yes"
    default_second = "yes"
    while True:
        user_input = input(
            "Do you want to run synthesis in incremental mode?"
            + "(options = yes/no) [default = yes]: "
        )
        if user_input == "yes" or user_input == "no" or user_input == "":
            user_input = default if user_input == "" else user_input
            break
    if stop_step is StopStep.SYNTHESIS:
        return IsIncremental(
            synthesis=True if user_input == "yes" else False,
            implementation=True,
        )
    while True:
        user_second_input = input(
            "Do you want to run implementation in incremental mode?"
            + "(options = yes/no) [default = yes]:"
        )
        if (
            user_second_input == "yes"
            or user_second_input == "no"
            or user_second_input == ""
        ):
            user_second_input = (
                default_second
                if user_second_input == ""
                else user_second_input
            )
            return IsIncremental(
                synthesis=True if user_input == "yes" else False,
                implementation=True if user_second_input == "yes" else False,
            )


def _ask_implementation_directives(
    to_request: ImplementationStep, incremental_mode: IsIncremental,
) -> str:
    default = "Default"
    directives_paragraph = doc.get_directives_paragraph(
        to_request.name.lower()
    )
    directives = doc.get_directives(directives_paragraph)
    incremental_directives = doc.get_directives(
        doc.get_directives_paragraph("read checkpoint")
    )
    while True:
        user_input = input(
            "What directive do you want to give to the "
            + str(to_request)
            + ' step? Input"?"'
            + " to get help on available directives."
            + " [default = "
            + default
            + "]: "
        )
        if user_input == "?":
            if incremental_mode.implementation:
                print(
                    "Available directives are only the ones compatible"
                    + " with read_checkpoint thus: "
                    + str(incremental_directives)
                )
            else:
                print("Available directives: " + directives_paragraph)
            continue
        if (
            user_input
            in (
                incremental_directives
                if incremental_mode.implementation
                else directives
            )
            or user_input == ""
        ):
            return default if user_input == "" else user_input


def ask_directives(
    stop_step: StopStep, incremental_mode: IsIncremental
) -> Tuple[str, Optional[str], Optional[str]]:
    synth_directives_paragraph = doc.get_directives_paragraph("synthesis")
    synth_directives = doc.get_directives(synth_directives_paragraph)
    or_directives = ""
    default = "default"
    for i in synth_directives:
        or_directives += i + "|"
    or_directives = or_directives[: len(or_directives) - 1]
    while True:
        user_input = input(
            "What directives do you want to give to the"
            + ' synthesis step? Input "?" to get help on'
            + " available directives."
            + " [default = "
            + default
            + "]: "
        )
        if user_input == "?":
            print(synth_directives_paragraph)
            continue
        if user_input == "":
            user_input = default
        if user_input in synth_directives:
            input_synth_directives = user_input
            break
    if stop_step is StopStep.SYNTHESIS:
        return input_synth_directives, None, None
    else:
        return (
            input_synth_directives,
            _ask_implementation_directives(
                ImplementationStep.PLACE, incremental_mode
            ),
            _ask_implementation_directives(
                ImplementationStep.ROUTE, incremental_mode
            ),
        )


def ask_clock() -> int:
    default = "250"
    while True:
        user_input = input(
            "Input target clock as an integer value (in Mhz) [default: "
            + default
            + "]: "
        )
        if user_input == "":
            user_input = default
        try:
            return int(user_input)
        except ValueError:
            print("Invalid, input\n")
            continue


def is_valid_clock(
    parsed_src: SourceParser, port: Optional[Port]
) -> Tuple[Optional[Port], Optional[PortDirection], Optional[PortType]]:
    if not port:
        return None, None, None
    port_direction = port.get_direction()
    port_type = port.get_type()
    if (
        port_direction.direction is PortDirectionEnum.INPUT
        and port_type.type is PortTypeEnum.SCALAR
        if parsed_src.get_hdl() is RTL.VHDL
        else port_type.type is PortTypeEnum.VECTOR
    ):
        return (
            port,
            port_direction,
            port_type,
        )
    return None, port_direction, port_type


def ask_identifiers(parsed_src: SourceParser) -> Port:
    while True:
        user_input = input("Enter clock identifier: ")
        clock, clock_direction, clock_port_type = is_valid_clock(
            parsed_src, parsed_src.check_port(user_input)
        )
        if clock:
            return clock
        print(
            "Invalid clock identifier: port direction = "
            + str(clock_direction)
            + " (must be an input port) port type = "
            + str(clock_port_type)
            + " (must be binary)\n"
            + "available ports: "
            + str(parsed_src.get_ports())
        )


def ask_parameters_range(
    param_list: List[Parameter], config: Configuration
) -> "OrderedDict[str, Tuple[int, int]]":
    shortcut = "_"
    print(
        "For each parameter specify a range e.g.'0 100' ,\n"
        + "a shorthand for the maximum integer value is int_high so "
        + "'-int_high int_high' specifies the whole integer range while "
        + "'0 int_high' specifies a natural for example. \n"
        + "Enter "
        + shortcut
        + " at any point to fix all unspecified ranges to "
        + "'-int_high int_high'"
    )
    param_range_accumulator = []
    i = 0
    while i < len(param_list):
        user_input = input(
            "Enter range for " + str(param_list[i].get_name()) + ": "
        )
        if user_input == shortcut:
            return OrderedDict(
                zip(
                    [p.get_name() for p in param_list],
                    param_range_accumulator
                    + (
                        [
                            (
                                -int(config.get_config("INTEGER_HIGH")),
                                config.get_config("INTEGER_HIGH"),
                            )
                        ]
                        * (len(param_list) - len(param_range_accumulator))
                    ),
                )
            )
        splitted_input = user_input.strip().split()
        if len(splitted_input) > 2:
            print("Please input only two integers separated by a whitespace")
            continue
        try:
            param_range_accumulator.append(
                (int(splitted_input[0]), int(splitted_input[1]))
            )
            i = i + 1
        except Exception:
            print("Could not parse input as a pair of integers ")

    return OrderedDict(
        zip([p.get_name() for p in param_list], param_range_accumulator)
    )


def ask_parameters(parsed_src: SourceParser) -> List[Parameter]:
    while True:
        user_input = input(
            "Enter a comma separated list "
            + "of parameters to optimize [example: firstParam, N, width]: "
        )
        param_list = parse_comma_separated_list(r"\w+", user_input)
        available_param_list = [i for i in parsed_src.get_parameters()]
        if param_list and all(
            i in [p.get_name() for p in available_param_list]
            for i in param_list
        ):
            return [
                parsed_src.get_parameter(i)
                for i in [p.get_name() for p in available_param_list]
                if i in param_list
            ]  # return ordered list
        if not param_list:
            print("Malformed list: " + user_input + "\ntry again\n")
        else:
            print(
                "The following parameters could not be found among "
                + "the ones declared in "
                + parsed_src.get_path()
                + " at module "
                + parsed_src.get_top_entity().get_name()
                + ": "
                + str([i for i in param_list if i not in available_param_list])
                + "\nAvailable parameters are "
                + str(available_param_list)
                + "\n"
            )


def parse_comma_separated_list(
    regexp_element: str, to_parse: str
) -> Optional[List[str]]:
    return (
        None
        if not re.fullmatch(
            "^("
            + regexp_element
            + ")"
            + "(([ \t]*,[ \t]*)("
            + regexp_element
            + "))*$",
            to_parse,
        )
        else [i for i in re.sub(r"[ \t]*", "", to_parse).split(",")]
    )


def ask_utilization_metrics(
    util_indices: Dict[str, List[str]]
) -> List[Metric]:

    info_prompt = "Percentage Utilisation indices available:\n"
    i = 0
    counter_dict = {}
    for section in util_indices.keys():
        info_prompt += section + "\n"
        for util_metric in util_indices[section]:
            i = i + 1
            counter_dict[i] = Metric(
                utilisation=(section, util_metric), is_frequency=False
            )
            info_prompt += "\t(" + str(i) + ") " + util_metric + "\n"

    print(info_prompt)
    default = "-1, 1, 4, 9"
    while True:
        user_input = input(
            "Enter a comma-separated list of percentage "
            + "indices to optimize, input -1 to select frequency [default = "
            + default
            + "]: "
        )
        if user_input == "":
            user_input = default
        try:
            parsed_list = [
                int(i)
                for i in parse_comma_separated_list(r"[-+]?\d+", user_input)
            ]
            if parsed_list and max([i for i in parsed_list if i > 0]) < len(
                counter_dict
            ):
                out: List[Metric] = [
                    counter_dict[i]
                    for i in set([i for i in parsed_list if i >= 0])
                ]
                return (
                    (out + [Metric(None, True)])
                    if (-1 in parsed_list)
                    else out
                )
        except Exception as e:
            print(e)
            print(
                "Invalid input, please enter"
                + " a comma separated list of integer numbers"
            )


def ask_is_point_evaluation() -> bool:
    default = "p"
    while True:
        user_input = input(
            "Would you like to run single point evaluation "
            + "or design space evaluation?\n"
            + "Please input 'p' for point evaluation "
            + "or 'd' for design space evaluation "
            "[default = " + default + "]: "
        )
        user_input = user_input.strip()
        if not (user_input == "p" or user_input == "d"):
            print(
                user_input
                + " is an in invalid input, please input either 'p' or 'd'"
            )
            continue
        return True if user_input == "p" else False


def ask_parameters_value(param_list: List[Parameter]) -> List[int]:
    if len(param_list) == 0:
        raise Exception(
            "No params passed to ask_parameters_value, at least one must be selected"
        )
    parameter_value_accumulator = []
    i = 0
    while i < len(param_list):
        user_input = input(
            "Enter value for parameter " + param_list[i].get_name() + ": "
        )
        try:
            parameter_value_accumulator.append(int(user_input))
            i = i + 1
        except Exception:
            print("Please input an integer value")
            continue
    return parameter_value_accumulator



# Generated from Verilog2001.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .Verilog2001Parser import Verilog2001Parser
else:
    from Verilog2001Parser import Verilog2001Parser

# This class defines a complete generic visitor for a parse tree produced by Verilog2001Parser.

class Verilog2001Visitor(ParseTreeVisitor):

    # Visit a parse tree produced by Verilog2001Parser#config_declaration.
    def visitConfig_declaration(self, ctx:Verilog2001Parser.Config_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#design_statement.
    def visitDesign_statement(self, ctx:Verilog2001Parser.Design_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#config_rule_statement.
    def visitConfig_rule_statement(self, ctx:Verilog2001Parser.Config_rule_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#default_clause.
    def visitDefault_clause(self, ctx:Verilog2001Parser.Default_clauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#inst_clause.
    def visitInst_clause(self, ctx:Verilog2001Parser.Inst_clauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#inst_name.
    def visitInst_name(self, ctx:Verilog2001Parser.Inst_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#liblist_clause.
    def visitLiblist_clause(self, ctx:Verilog2001Parser.Liblist_clauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#cell_clause.
    def visitCell_clause(self, ctx:Verilog2001Parser.Cell_clauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#use_clause.
    def visitUse_clause(self, ctx:Verilog2001Parser.Use_clauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#source_text.
    def visitSource_text(self, ctx:Verilog2001Parser.Source_textContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#description.
    def visitDescription(self, ctx:Verilog2001Parser.DescriptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_declaration.
    def visitModule_declaration(self, ctx:Verilog2001Parser.Module_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_keyword.
    def visitModule_keyword(self, ctx:Verilog2001Parser.Module_keywordContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_parameter_port_list.
    def visitModule_parameter_port_list(self, ctx:Verilog2001Parser.Module_parameter_port_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_ports.
    def visitList_of_ports(self, ctx:Verilog2001Parser.List_of_portsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_port_declarations.
    def visitList_of_port_declarations(self, ctx:Verilog2001Parser.List_of_port_declarationsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#port.
    def visitPort(self, ctx:Verilog2001Parser.PortContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#port_expression.
    def visitPort_expression(self, ctx:Verilog2001Parser.Port_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#port_reference.
    def visitPort_reference(self, ctx:Verilog2001Parser.Port_referenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#port_declaration.
    def visitPort_declaration(self, ctx:Verilog2001Parser.Port_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_item.
    def visitModule_item(self, ctx:Verilog2001Parser.Module_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_or_generate_item.
    def visitModule_or_generate_item(self, ctx:Verilog2001Parser.Module_or_generate_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#non_port_module_item.
    def visitNon_port_module_item(self, ctx:Verilog2001Parser.Non_port_module_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_or_generate_item_declaration.
    def visitModule_or_generate_item_declaration(self, ctx:Verilog2001Parser.Module_or_generate_item_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#parameter_override.
    def visitParameter_override(self, ctx:Verilog2001Parser.Parameter_overrideContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#local_parameter_declaration.
    def visitLocal_parameter_declaration(self, ctx:Verilog2001Parser.Local_parameter_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#parameter_declaration.
    def visitParameter_declaration(self, ctx:Verilog2001Parser.Parameter_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#parameter_declaration_.
    def visitParameter_declaration_(self, ctx:Verilog2001Parser.Parameter_declaration_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#specparam_declaration.
    def visitSpecparam_declaration(self, ctx:Verilog2001Parser.Specparam_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#inout_declaration.
    def visitInout_declaration(self, ctx:Verilog2001Parser.Inout_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#input_declaration.
    def visitInput_declaration(self, ctx:Verilog2001Parser.Input_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#output_declaration.
    def visitOutput_declaration(self, ctx:Verilog2001Parser.Output_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#event_declaration.
    def visitEvent_declaration(self, ctx:Verilog2001Parser.Event_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#genvar_declaration.
    def visitGenvar_declaration(self, ctx:Verilog2001Parser.Genvar_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#integer_declaration.
    def visitInteger_declaration(self, ctx:Verilog2001Parser.Integer_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#time_declaration.
    def visitTime_declaration(self, ctx:Verilog2001Parser.Time_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#real_declaration.
    def visitReal_declaration(self, ctx:Verilog2001Parser.Real_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#realtime_declaration.
    def visitRealtime_declaration(self, ctx:Verilog2001Parser.Realtime_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#reg_declaration.
    def visitReg_declaration(self, ctx:Verilog2001Parser.Reg_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#net_declaration.
    def visitNet_declaration(self, ctx:Verilog2001Parser.Net_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#net_type.
    def visitNet_type(self, ctx:Verilog2001Parser.Net_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#output_variable_type.
    def visitOutput_variable_type(self, ctx:Verilog2001Parser.Output_variable_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#real_type.
    def visitReal_type(self, ctx:Verilog2001Parser.Real_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#variable_type.
    def visitVariable_type(self, ctx:Verilog2001Parser.Variable_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#drive_strength.
    def visitDrive_strength(self, ctx:Verilog2001Parser.Drive_strengthContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#strength0.
    def visitStrength0(self, ctx:Verilog2001Parser.Strength0Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#strength1.
    def visitStrength1(self, ctx:Verilog2001Parser.Strength1Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#charge_strength.
    def visitCharge_strength(self, ctx:Verilog2001Parser.Charge_strengthContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#delay3.
    def visitDelay3(self, ctx:Verilog2001Parser.Delay3Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#delay2.
    def visitDelay2(self, ctx:Verilog2001Parser.Delay2Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#delay_value.
    def visitDelay_value(self, ctx:Verilog2001Parser.Delay_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_event_identifiers.
    def visitList_of_event_identifiers(self, ctx:Verilog2001Parser.List_of_event_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_net_identifiers.
    def visitList_of_net_identifiers(self, ctx:Verilog2001Parser.List_of_net_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_genvar_identifiers.
    def visitList_of_genvar_identifiers(self, ctx:Verilog2001Parser.List_of_genvar_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_port_identifiers.
    def visitList_of_port_identifiers(self, ctx:Verilog2001Parser.List_of_port_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_net_decl_assignments.
    def visitList_of_net_decl_assignments(self, ctx:Verilog2001Parser.List_of_net_decl_assignmentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_param_assignments.
    def visitList_of_param_assignments(self, ctx:Verilog2001Parser.List_of_param_assignmentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_specparam_assignments.
    def visitList_of_specparam_assignments(self, ctx:Verilog2001Parser.List_of_specparam_assignmentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_real_identifiers.
    def visitList_of_real_identifiers(self, ctx:Verilog2001Parser.List_of_real_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_variable_identifiers.
    def visitList_of_variable_identifiers(self, ctx:Verilog2001Parser.List_of_variable_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_variable_port_identifiers.
    def visitList_of_variable_port_identifiers(self, ctx:Verilog2001Parser.List_of_variable_port_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#net_decl_assignment.
    def visitNet_decl_assignment(self, ctx:Verilog2001Parser.Net_decl_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#param_assignment.
    def visitParam_assignment(self, ctx:Verilog2001Parser.Param_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#specparam_assignment.
    def visitSpecparam_assignment(self, ctx:Verilog2001Parser.Specparam_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#pulse_control_specparam.
    def visitPulse_control_specparam(self, ctx:Verilog2001Parser.Pulse_control_specparamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#error_limit_value.
    def visitError_limit_value(self, ctx:Verilog2001Parser.Error_limit_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#reject_limit_value.
    def visitReject_limit_value(self, ctx:Verilog2001Parser.Reject_limit_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#limit_value.
    def visitLimit_value(self, ctx:Verilog2001Parser.Limit_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#dimension.
    def visitDimension(self, ctx:Verilog2001Parser.DimensionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#range_.
    def visitRange_(self, ctx:Verilog2001Parser.Range_Context):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_declaration.
    def visitFunction_declaration(self, ctx:Verilog2001Parser.Function_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_item_declaration.
    def visitFunction_item_declaration(self, ctx:Verilog2001Parser.Function_item_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_port_list.
    def visitFunction_port_list(self, ctx:Verilog2001Parser.Function_port_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_port.
    def visitFunction_port(self, ctx:Verilog2001Parser.Function_portContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#range_or_type.
    def visitRange_or_type(self, ctx:Verilog2001Parser.Range_or_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#task_declaration.
    def visitTask_declaration(self, ctx:Verilog2001Parser.Task_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#task_item_declaration.
    def visitTask_item_declaration(self, ctx:Verilog2001Parser.Task_item_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#task_port_list.
    def visitTask_port_list(self, ctx:Verilog2001Parser.Task_port_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#task_port_item.
    def visitTask_port_item(self, ctx:Verilog2001Parser.Task_port_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#tf_decl_header.
    def visitTf_decl_header(self, ctx:Verilog2001Parser.Tf_decl_headerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#tf_declaration.
    def visitTf_declaration(self, ctx:Verilog2001Parser.Tf_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#task_port_type.
    def visitTask_port_type(self, ctx:Verilog2001Parser.Task_port_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#block_item_declaration.
    def visitBlock_item_declaration(self, ctx:Verilog2001Parser.Block_item_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#block_reg_declaration.
    def visitBlock_reg_declaration(self, ctx:Verilog2001Parser.Block_reg_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_block_variable_identifiers.
    def visitList_of_block_variable_identifiers(self, ctx:Verilog2001Parser.List_of_block_variable_identifiersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#block_variable_type.
    def visitBlock_variable_type(self, ctx:Verilog2001Parser.Block_variable_typeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#gate_instantiation.
    def visitGate_instantiation(self, ctx:Verilog2001Parser.Gate_instantiationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#cmos_switch_instance.
    def visitCmos_switch_instance(self, ctx:Verilog2001Parser.Cmos_switch_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#enable_gate_instance.
    def visitEnable_gate_instance(self, ctx:Verilog2001Parser.Enable_gate_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#mos_switch_instance.
    def visitMos_switch_instance(self, ctx:Verilog2001Parser.Mos_switch_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#n_input_gate_instance.
    def visitN_input_gate_instance(self, ctx:Verilog2001Parser.N_input_gate_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#n_output_gate_instance.
    def visitN_output_gate_instance(self, ctx:Verilog2001Parser.N_output_gate_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#pass_switch_instance.
    def visitPass_switch_instance(self, ctx:Verilog2001Parser.Pass_switch_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#pass_enable_switch_instance.
    def visitPass_enable_switch_instance(self, ctx:Verilog2001Parser.Pass_enable_switch_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#pull_gate_instance.
    def visitPull_gate_instance(self, ctx:Verilog2001Parser.Pull_gate_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#name_of_gate_instance.
    def visitName_of_gate_instance(self, ctx:Verilog2001Parser.Name_of_gate_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#pulldown_strength.
    def visitPulldown_strength(self, ctx:Verilog2001Parser.Pulldown_strengthContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#pullup_strength.
    def visitPullup_strength(self, ctx:Verilog2001Parser.Pullup_strengthContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#enable_terminal.
    def visitEnable_terminal(self, ctx:Verilog2001Parser.Enable_terminalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#ncontrol_terminal.
    def visitNcontrol_terminal(self, ctx:Verilog2001Parser.Ncontrol_terminalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#pcontrol_terminal.
    def visitPcontrol_terminal(self, ctx:Verilog2001Parser.Pcontrol_terminalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#input_terminal.
    def visitInput_terminal(self, ctx:Verilog2001Parser.Input_terminalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#inout_terminal.
    def visitInout_terminal(self, ctx:Verilog2001Parser.Inout_terminalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#output_terminal.
    def visitOutput_terminal(self, ctx:Verilog2001Parser.Output_terminalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#cmos_switchtype.
    def visitCmos_switchtype(self, ctx:Verilog2001Parser.Cmos_switchtypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#enable_gatetype.
    def visitEnable_gatetype(self, ctx:Verilog2001Parser.Enable_gatetypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#mos_switchtype.
    def visitMos_switchtype(self, ctx:Verilog2001Parser.Mos_switchtypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#n_input_gatetype.
    def visitN_input_gatetype(self, ctx:Verilog2001Parser.N_input_gatetypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#n_output_gatetype.
    def visitN_output_gatetype(self, ctx:Verilog2001Parser.N_output_gatetypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#pass_en_switchtype.
    def visitPass_en_switchtype(self, ctx:Verilog2001Parser.Pass_en_switchtypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#pass_switchtype.
    def visitPass_switchtype(self, ctx:Verilog2001Parser.Pass_switchtypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_instantiation.
    def visitModule_instantiation(self, ctx:Verilog2001Parser.Module_instantiationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#parameter_value_assignment.
    def visitParameter_value_assignment(self, ctx:Verilog2001Parser.Parameter_value_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_parameter_assignments.
    def visitList_of_parameter_assignments(self, ctx:Verilog2001Parser.List_of_parameter_assignmentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#ordered_parameter_assignment.
    def visitOrdered_parameter_assignment(self, ctx:Verilog2001Parser.Ordered_parameter_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#named_parameter_assignment.
    def visitNamed_parameter_assignment(self, ctx:Verilog2001Parser.Named_parameter_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_instance.
    def visitModule_instance(self, ctx:Verilog2001Parser.Module_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#name_of_instance.
    def visitName_of_instance(self, ctx:Verilog2001Parser.Name_of_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_port_connections.
    def visitList_of_port_connections(self, ctx:Verilog2001Parser.List_of_port_connectionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#ordered_port_connection.
    def visitOrdered_port_connection(self, ctx:Verilog2001Parser.Ordered_port_connectionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#named_port_connection.
    def visitNamed_port_connection(self, ctx:Verilog2001Parser.Named_port_connectionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#generated_instantiation.
    def visitGenerated_instantiation(self, ctx:Verilog2001Parser.Generated_instantiationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#generate_item_or_null.
    def visitGenerate_item_or_null(self, ctx:Verilog2001Parser.Generate_item_or_nullContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#generate_item.
    def visitGenerate_item(self, ctx:Verilog2001Parser.Generate_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#generate_conditional_statement.
    def visitGenerate_conditional_statement(self, ctx:Verilog2001Parser.Generate_conditional_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#generate_case_statement.
    def visitGenerate_case_statement(self, ctx:Verilog2001Parser.Generate_case_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#genvar_case_item.
    def visitGenvar_case_item(self, ctx:Verilog2001Parser.Genvar_case_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#generate_loop_statement.
    def visitGenerate_loop_statement(self, ctx:Verilog2001Parser.Generate_loop_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#genvar_assignment.
    def visitGenvar_assignment(self, ctx:Verilog2001Parser.Genvar_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#generate_block.
    def visitGenerate_block(self, ctx:Verilog2001Parser.Generate_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#continuous_assign.
    def visitContinuous_assign(self, ctx:Verilog2001Parser.Continuous_assignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_net_assignments.
    def visitList_of_net_assignments(self, ctx:Verilog2001Parser.List_of_net_assignmentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#net_assignment.
    def visitNet_assignment(self, ctx:Verilog2001Parser.Net_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#initial_construct.
    def visitInitial_construct(self, ctx:Verilog2001Parser.Initial_constructContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#always_construct.
    def visitAlways_construct(self, ctx:Verilog2001Parser.Always_constructContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#blocking_assignment.
    def visitBlocking_assignment(self, ctx:Verilog2001Parser.Blocking_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#nonblocking_assignment.
    def visitNonblocking_assignment(self, ctx:Verilog2001Parser.Nonblocking_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#procedural_continuous_assignments.
    def visitProcedural_continuous_assignments(self, ctx:Verilog2001Parser.Procedural_continuous_assignmentsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_blocking_assignment.
    def visitFunction_blocking_assignment(self, ctx:Verilog2001Parser.Function_blocking_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_statement_or_null.
    def visitFunction_statement_or_null(self, ctx:Verilog2001Parser.Function_statement_or_nullContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_seq_block.
    def visitFunction_seq_block(self, ctx:Verilog2001Parser.Function_seq_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#variable_assignment.
    def visitVariable_assignment(self, ctx:Verilog2001Parser.Variable_assignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#par_block.
    def visitPar_block(self, ctx:Verilog2001Parser.Par_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#seq_block.
    def visitSeq_block(self, ctx:Verilog2001Parser.Seq_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#statement.
    def visitStatement(self, ctx:Verilog2001Parser.StatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#statement_or_null.
    def visitStatement_or_null(self, ctx:Verilog2001Parser.Statement_or_nullContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_statement.
    def visitFunction_statement(self, ctx:Verilog2001Parser.Function_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#delay_or_event_control.
    def visitDelay_or_event_control(self, ctx:Verilog2001Parser.Delay_or_event_controlContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#delay_control.
    def visitDelay_control(self, ctx:Verilog2001Parser.Delay_controlContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#disable_statement.
    def visitDisable_statement(self, ctx:Verilog2001Parser.Disable_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#event_control.
    def visitEvent_control(self, ctx:Verilog2001Parser.Event_controlContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#event_trigger.
    def visitEvent_trigger(self, ctx:Verilog2001Parser.Event_triggerContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#event_expression.
    def visitEvent_expression(self, ctx:Verilog2001Parser.Event_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#event_primary.
    def visitEvent_primary(self, ctx:Verilog2001Parser.Event_primaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#procedural_timing_control_statement.
    def visitProcedural_timing_control_statement(self, ctx:Verilog2001Parser.Procedural_timing_control_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#wait_statement.
    def visitWait_statement(self, ctx:Verilog2001Parser.Wait_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#conditional_statement.
    def visitConditional_statement(self, ctx:Verilog2001Parser.Conditional_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#if_else_if_statement.
    def visitIf_else_if_statement(self, ctx:Verilog2001Parser.If_else_if_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_conditional_statement.
    def visitFunction_conditional_statement(self, ctx:Verilog2001Parser.Function_conditional_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_if_else_if_statement.
    def visitFunction_if_else_if_statement(self, ctx:Verilog2001Parser.Function_if_else_if_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#case_statement.
    def visitCase_statement(self, ctx:Verilog2001Parser.Case_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#case_item.
    def visitCase_item(self, ctx:Verilog2001Parser.Case_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_case_statement.
    def visitFunction_case_statement(self, ctx:Verilog2001Parser.Function_case_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_case_item.
    def visitFunction_case_item(self, ctx:Verilog2001Parser.Function_case_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_loop_statement.
    def visitFunction_loop_statement(self, ctx:Verilog2001Parser.Function_loop_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#loop_statement.
    def visitLoop_statement(self, ctx:Verilog2001Parser.Loop_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#system_task_enable.
    def visitSystem_task_enable(self, ctx:Verilog2001Parser.System_task_enableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#task_enable.
    def visitTask_enable(self, ctx:Verilog2001Parser.Task_enableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#specify_block.
    def visitSpecify_block(self, ctx:Verilog2001Parser.Specify_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#specify_item.
    def visitSpecify_item(self, ctx:Verilog2001Parser.Specify_itemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#pulsestyle_declaration.
    def visitPulsestyle_declaration(self, ctx:Verilog2001Parser.Pulsestyle_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#showcancelled_declaration.
    def visitShowcancelled_declaration(self, ctx:Verilog2001Parser.Showcancelled_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#path_declaration.
    def visitPath_declaration(self, ctx:Verilog2001Parser.Path_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#simple_path_declaration.
    def visitSimple_path_declaration(self, ctx:Verilog2001Parser.Simple_path_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#parallel_path_description.
    def visitParallel_path_description(self, ctx:Verilog2001Parser.Parallel_path_descriptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#full_path_description.
    def visitFull_path_description(self, ctx:Verilog2001Parser.Full_path_descriptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_path_inputs.
    def visitList_of_path_inputs(self, ctx:Verilog2001Parser.List_of_path_inputsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_path_outputs.
    def visitList_of_path_outputs(self, ctx:Verilog2001Parser.List_of_path_outputsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#specify_input_terminal_descriptor.
    def visitSpecify_input_terminal_descriptor(self, ctx:Verilog2001Parser.Specify_input_terminal_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#specify_output_terminal_descriptor.
    def visitSpecify_output_terminal_descriptor(self, ctx:Verilog2001Parser.Specify_output_terminal_descriptorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#input_identifier.
    def visitInput_identifier(self, ctx:Verilog2001Parser.Input_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#output_identifier.
    def visitOutput_identifier(self, ctx:Verilog2001Parser.Output_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#path_delay_value.
    def visitPath_delay_value(self, ctx:Verilog2001Parser.Path_delay_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#list_of_path_delay_expressions.
    def visitList_of_path_delay_expressions(self, ctx:Verilog2001Parser.List_of_path_delay_expressionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#t_path_delay_expression.
    def visitT_path_delay_expression(self, ctx:Verilog2001Parser.T_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#trise_path_delay_expression.
    def visitTrise_path_delay_expression(self, ctx:Verilog2001Parser.Trise_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#tfall_path_delay_expression.
    def visitTfall_path_delay_expression(self, ctx:Verilog2001Parser.Tfall_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#tz_path_delay_expression.
    def visitTz_path_delay_expression(self, ctx:Verilog2001Parser.Tz_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#t01_path_delay_expression.
    def visitT01_path_delay_expression(self, ctx:Verilog2001Parser.T01_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#t10_path_delay_expression.
    def visitT10_path_delay_expression(self, ctx:Verilog2001Parser.T10_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#t0z_path_delay_expression.
    def visitT0z_path_delay_expression(self, ctx:Verilog2001Parser.T0z_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#tz1_path_delay_expression.
    def visitTz1_path_delay_expression(self, ctx:Verilog2001Parser.Tz1_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#t1z_path_delay_expression.
    def visitT1z_path_delay_expression(self, ctx:Verilog2001Parser.T1z_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#tz0_path_delay_expression.
    def visitTz0_path_delay_expression(self, ctx:Verilog2001Parser.Tz0_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#t0x_path_delay_expression.
    def visitT0x_path_delay_expression(self, ctx:Verilog2001Parser.T0x_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#tx1_path_delay_expression.
    def visitTx1_path_delay_expression(self, ctx:Verilog2001Parser.Tx1_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#t1x_path_delay_expression.
    def visitT1x_path_delay_expression(self, ctx:Verilog2001Parser.T1x_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#tx0_path_delay_expression.
    def visitTx0_path_delay_expression(self, ctx:Verilog2001Parser.Tx0_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#txz_path_delay_expression.
    def visitTxz_path_delay_expression(self, ctx:Verilog2001Parser.Txz_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#tzx_path_delay_expression.
    def visitTzx_path_delay_expression(self, ctx:Verilog2001Parser.Tzx_path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#path_delay_expression.
    def visitPath_delay_expression(self, ctx:Verilog2001Parser.Path_delay_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#edge_sensitive_path_declaration.
    def visitEdge_sensitive_path_declaration(self, ctx:Verilog2001Parser.Edge_sensitive_path_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#parallel_edge_sensitive_path_description.
    def visitParallel_edge_sensitive_path_description(self, ctx:Verilog2001Parser.Parallel_edge_sensitive_path_descriptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#full_edge_sensitive_path_description.
    def visitFull_edge_sensitive_path_description(self, ctx:Verilog2001Parser.Full_edge_sensitive_path_descriptionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#data_source_expression.
    def visitData_source_expression(self, ctx:Verilog2001Parser.Data_source_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#edge_identifier.
    def visitEdge_identifier(self, ctx:Verilog2001Parser.Edge_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#state_dependent_path_declaration.
    def visitState_dependent_path_declaration(self, ctx:Verilog2001Parser.State_dependent_path_declarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#polarity_operator.
    def visitPolarity_operator(self, ctx:Verilog2001Parser.Polarity_operatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#checktime_condition.
    def visitChecktime_condition(self, ctx:Verilog2001Parser.Checktime_conditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#delayed_data.
    def visitDelayed_data(self, ctx:Verilog2001Parser.Delayed_dataContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#delayed_reference.
    def visitDelayed_reference(self, ctx:Verilog2001Parser.Delayed_referenceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#end_edge_offset.
    def visitEnd_edge_offset(self, ctx:Verilog2001Parser.End_edge_offsetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#event_based_flag.
    def visitEvent_based_flag(self, ctx:Verilog2001Parser.Event_based_flagContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#notify_reg.
    def visitNotify_reg(self, ctx:Verilog2001Parser.Notify_regContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#remain_active_flag.
    def visitRemain_active_flag(self, ctx:Verilog2001Parser.Remain_active_flagContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#stamptime_condition.
    def visitStamptime_condition(self, ctx:Verilog2001Parser.Stamptime_conditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#start_edge_offset.
    def visitStart_edge_offset(self, ctx:Verilog2001Parser.Start_edge_offsetContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#threshold.
    def visitThreshold(self, ctx:Verilog2001Parser.ThresholdContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#timing_check_limit.
    def visitTiming_check_limit(self, ctx:Verilog2001Parser.Timing_check_limitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#concatenation.
    def visitConcatenation(self, ctx:Verilog2001Parser.ConcatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#constant_concatenation.
    def visitConstant_concatenation(self, ctx:Verilog2001Parser.Constant_concatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#constant_multiple_concatenation.
    def visitConstant_multiple_concatenation(self, ctx:Verilog2001Parser.Constant_multiple_concatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_path_concatenation.
    def visitModule_path_concatenation(self, ctx:Verilog2001Parser.Module_path_concatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_path_multiple_concatenation.
    def visitModule_path_multiple_concatenation(self, ctx:Verilog2001Parser.Module_path_multiple_concatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#multiple_concatenation.
    def visitMultiple_concatenation(self, ctx:Verilog2001Parser.Multiple_concatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#net_concatenation.
    def visitNet_concatenation(self, ctx:Verilog2001Parser.Net_concatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#net_concatenation_value.
    def visitNet_concatenation_value(self, ctx:Verilog2001Parser.Net_concatenation_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#variable_concatenation.
    def visitVariable_concatenation(self, ctx:Verilog2001Parser.Variable_concatenationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#variable_concatenation_value.
    def visitVariable_concatenation_value(self, ctx:Verilog2001Parser.Variable_concatenation_valueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#constant_function_call.
    def visitConstant_function_call(self, ctx:Verilog2001Parser.Constant_function_callContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_call.
    def visitFunction_call(self, ctx:Verilog2001Parser.Function_callContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#system_function_call.
    def visitSystem_function_call(self, ctx:Verilog2001Parser.System_function_callContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#genvar_function_call.
    def visitGenvar_function_call(self, ctx:Verilog2001Parser.Genvar_function_callContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#base_expression.
    def visitBase_expression(self, ctx:Verilog2001Parser.Base_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#constant_base_expression.
    def visitConstant_base_expression(self, ctx:Verilog2001Parser.Constant_base_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#constant_expression.
    def visitConstant_expression(self, ctx:Verilog2001Parser.Constant_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#constant_mintypmax_expression.
    def visitConstant_mintypmax_expression(self, ctx:Verilog2001Parser.Constant_mintypmax_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#constant_range_expression.
    def visitConstant_range_expression(self, ctx:Verilog2001Parser.Constant_range_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#dimension_constant_expression.
    def visitDimension_constant_expression(self, ctx:Verilog2001Parser.Dimension_constant_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#expression.
    def visitExpression(self, ctx:Verilog2001Parser.ExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#term.
    def visitTerm(self, ctx:Verilog2001Parser.TermContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#lsb_constant_expression.
    def visitLsb_constant_expression(self, ctx:Verilog2001Parser.Lsb_constant_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#mintypmax_expression.
    def visitMintypmax_expression(self, ctx:Verilog2001Parser.Mintypmax_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_path_conditional_expression.
    def visitModule_path_conditional_expression(self, ctx:Verilog2001Parser.Module_path_conditional_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_path_expression.
    def visitModule_path_expression(self, ctx:Verilog2001Parser.Module_path_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_path_mintypmax_expression.
    def visitModule_path_mintypmax_expression(self, ctx:Verilog2001Parser.Module_path_mintypmax_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#msb_constant_expression.
    def visitMsb_constant_expression(self, ctx:Verilog2001Parser.Msb_constant_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#range_expression.
    def visitRange_expression(self, ctx:Verilog2001Parser.Range_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#width_constant_expression.
    def visitWidth_constant_expression(self, ctx:Verilog2001Parser.Width_constant_expressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#constant_primary.
    def visitConstant_primary(self, ctx:Verilog2001Parser.Constant_primaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_path_primary.
    def visitModule_path_primary(self, ctx:Verilog2001Parser.Module_path_primaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#primary.
    def visitPrimary(self, ctx:Verilog2001Parser.PrimaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#net_lvalue.
    def visitNet_lvalue(self, ctx:Verilog2001Parser.Net_lvalueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#variable_lvalue.
    def visitVariable_lvalue(self, ctx:Verilog2001Parser.Variable_lvalueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#unary_operator.
    def visitUnary_operator(self, ctx:Verilog2001Parser.Unary_operatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#binary_operator.
    def visitBinary_operator(self, ctx:Verilog2001Parser.Binary_operatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#unary_module_path_operator.
    def visitUnary_module_path_operator(self, ctx:Verilog2001Parser.Unary_module_path_operatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#binary_module_path_operator.
    def visitBinary_module_path_operator(self, ctx:Verilog2001Parser.Binary_module_path_operatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#number.
    def visitNumber(self, ctx:Verilog2001Parser.NumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#timing_spec.
    def visitTiming_spec(self, ctx:Verilog2001Parser.Timing_specContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#attribute_instance.
    def visitAttribute_instance(self, ctx:Verilog2001Parser.Attribute_instanceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#attr_spec.
    def visitAttr_spec(self, ctx:Verilog2001Parser.Attr_specContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#attr_name.
    def visitAttr_name(self, ctx:Verilog2001Parser.Attr_nameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#arrayed_identifier.
    def visitArrayed_identifier(self, ctx:Verilog2001Parser.Arrayed_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#block_identifier.
    def visitBlock_identifier(self, ctx:Verilog2001Parser.Block_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#cell_identifier.
    def visitCell_identifier(self, ctx:Verilog2001Parser.Cell_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#config_identifier.
    def visitConfig_identifier(self, ctx:Verilog2001Parser.Config_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#escaped_arrayed_identifier.
    def visitEscaped_arrayed_identifier(self, ctx:Verilog2001Parser.Escaped_arrayed_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#escaped_hierarchical_identifier.
    def visitEscaped_hierarchical_identifier(self, ctx:Verilog2001Parser.Escaped_hierarchical_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#event_identifier.
    def visitEvent_identifier(self, ctx:Verilog2001Parser.Event_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#function_identifier.
    def visitFunction_identifier(self, ctx:Verilog2001Parser.Function_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#gate_instance_identifier.
    def visitGate_instance_identifier(self, ctx:Verilog2001Parser.Gate_instance_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#generate_block_identifier.
    def visitGenerate_block_identifier(self, ctx:Verilog2001Parser.Generate_block_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#genvar_function_identifier.
    def visitGenvar_function_identifier(self, ctx:Verilog2001Parser.Genvar_function_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#genvar_identifier.
    def visitGenvar_identifier(self, ctx:Verilog2001Parser.Genvar_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#hierarchical_block_identifier.
    def visitHierarchical_block_identifier(self, ctx:Verilog2001Parser.Hierarchical_block_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#hierarchical_event_identifier.
    def visitHierarchical_event_identifier(self, ctx:Verilog2001Parser.Hierarchical_event_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#hierarchical_function_identifier.
    def visitHierarchical_function_identifier(self, ctx:Verilog2001Parser.Hierarchical_function_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#hierarchical_identifier.
    def visitHierarchical_identifier(self, ctx:Verilog2001Parser.Hierarchical_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#hierarchical_net_identifier.
    def visitHierarchical_net_identifier(self, ctx:Verilog2001Parser.Hierarchical_net_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#hierarchical_variable_identifier.
    def visitHierarchical_variable_identifier(self, ctx:Verilog2001Parser.Hierarchical_variable_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#hierarchical_task_identifier.
    def visitHierarchical_task_identifier(self, ctx:Verilog2001Parser.Hierarchical_task_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#identifier.
    def visitIdentifier(self, ctx:Verilog2001Parser.IdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#inout_port_identifier.
    def visitInout_port_identifier(self, ctx:Verilog2001Parser.Inout_port_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#input_port_identifier.
    def visitInput_port_identifier(self, ctx:Verilog2001Parser.Input_port_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#instance_identifier.
    def visitInstance_identifier(self, ctx:Verilog2001Parser.Instance_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#library_identifier.
    def visitLibrary_identifier(self, ctx:Verilog2001Parser.Library_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#memory_identifier.
    def visitMemory_identifier(self, ctx:Verilog2001Parser.Memory_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_identifier.
    def visitModule_identifier(self, ctx:Verilog2001Parser.Module_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#module_instance_identifier.
    def visitModule_instance_identifier(self, ctx:Verilog2001Parser.Module_instance_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#net_identifier.
    def visitNet_identifier(self, ctx:Verilog2001Parser.Net_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#output_port_identifier.
    def visitOutput_port_identifier(self, ctx:Verilog2001Parser.Output_port_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#parameter_identifier.
    def visitParameter_identifier(self, ctx:Verilog2001Parser.Parameter_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#port_identifier.
    def visitPort_identifier(self, ctx:Verilog2001Parser.Port_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#real_identifier.
    def visitReal_identifier(self, ctx:Verilog2001Parser.Real_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#simple_arrayed_identifier.
    def visitSimple_arrayed_identifier(self, ctx:Verilog2001Parser.Simple_arrayed_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#simple_hierarchical_identifier.
    def visitSimple_hierarchical_identifier(self, ctx:Verilog2001Parser.Simple_hierarchical_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#specparam_identifier.
    def visitSpecparam_identifier(self, ctx:Verilog2001Parser.Specparam_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#system_function_identifier.
    def visitSystem_function_identifier(self, ctx:Verilog2001Parser.System_function_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#system_task_identifier.
    def visitSystem_task_identifier(self, ctx:Verilog2001Parser.System_task_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#task_identifier.
    def visitTask_identifier(self, ctx:Verilog2001Parser.Task_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#terminal_identifier.
    def visitTerminal_identifier(self, ctx:Verilog2001Parser.Terminal_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#text_macro_identifier.
    def visitText_macro_identifier(self, ctx:Verilog2001Parser.Text_macro_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#topmodule_identifier.
    def visitTopmodule_identifier(self, ctx:Verilog2001Parser.Topmodule_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#udp_identifier.
    def visitUdp_identifier(self, ctx:Verilog2001Parser.Udp_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#udp_instance_identifier.
    def visitUdp_instance_identifier(self, ctx:Verilog2001Parser.Udp_instance_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#variable_identifier.
    def visitVariable_identifier(self, ctx:Verilog2001Parser.Variable_identifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#simple_hierarchical_branch.
    def visitSimple_hierarchical_branch(self, ctx:Verilog2001Parser.Simple_hierarchical_branchContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Verilog2001Parser#escaped_hierarchical_branch.
    def visitEscaped_hierarchical_branch(self, ctx:Verilog2001Parser.Escaped_hierarchical_branchContext):
        return self.visitChildren(ctx)



del Verilog2001Parser


import math
from random import uniform
from functools import lru_cache
from heapq import nlargest
from dovado_rtl.simple_types import Metric, Example
from dovado_rtl.estimation import Estimator
from dovado_rtl.point_evaluation import DesignPointEvaluator
from dovado_rtl.config import Configuration
import numpy as np
from typing import Tuple, List, Optional

from dovado_rtl.abstract_classes import AbstractFitnessEvaluator


class FitnessEvaluator(AbstractFitnessEvaluator):
    def __init__(
        self,
        approximating: bool,
        nearest_distance: int,
        estimator: Estimator,
        evaluator: DesignPointEvaluator,
        config: Configuration,
    ):
        self.__nearest_distance = nearest_distance
        self.__approximating = approximating
        self.__threshold: int = 0
        self.__estimator: Estimator = estimator
        self.__evaluator: DesignPointEvaluator = evaluator
        self.__config: Configuration = config
        self.__set_threshold(self.__estimator.get_examples())
        self.__last_design_point: Optional[Tuple[int, ...]] = None

    @lru_cache()
    def fitness(self, design_point: Tuple[int, ...], metric: Metric):
        # design_point is a Tuple because lists are unhashable
        # and caching is allowed only with hashable parameters

        sample_distance = self.__nth_nearest_distance(
            list(design_point),
            self.__estimator.get_examples(),
            int(self.__nearest_distance),
        )
        print("Nth nearest distance: " + str(sample_distance))
        if (
            not self.__approximating
            or (sample_distance == 0)
            or (sample_distance > self.__threshold)
        ):
            print("Fitness calling Vivado directly")
            full_design_value = self.__evaluator.evaluate(design_point)
            if not full_design_value:
                raise Exception("Evaluator returned a None design value")
            design_value = full_design_value.value[metric]
            if design_point != self.__last_design_point:
                self.__estimator.add_example(
                    Example(list(design_point), full_design_value)
                )
                self.__set_threshold(self.__estimator.get_examples())
        else:
            print("Fitness estimating")
            design_value = self.__estimator.estimate(design_point, metric)
            if design_value is None:
                if metric.is_frequency:
                    design_value = uniform(
                        0, 10000
                    )  # guessing from 0 to 10 GHz
                else:
                    design_value = uniform(
                        0, 100
                    )  # gussing from 0 to 100% utilisation
                print(
                    "An empty prediction was retrieved from the estimator for "
                    + str(design_point)
                    + " on "
                    + str(metric)
                    + "setting it to arbitrarily bad value "
                    + str(design_value)
                )
        print("design_point: " + str(design_point))
        print("metric: " + str(metric))
        print("Design Value: " + str(design_value))
        self.__last_design_point = design_point
        return design_value

    @staticmethod
    def __distance(
        design_point_1: List[int], design_point_2: List[int]
    ) -> float:
        if len(design_point_1) != len(design_point_2):
            raise ValueError(
                "Mismatching design points: different dimensionality"
            )
        return math.sqrt(
            np.sum(
                np.power(
                    np.subtract(
                        np.array(design_point_1), np.array(design_point_2)
                    ),
                    np.array([2] * len(design_point_1)),
                )
            )
            / len(design_point_1)
        )

    def __nth_nearest_distance(
        self, design_point: List[int], examples: List[Example], n: int,
    ) -> float:
        distances = []
        for example in examples:
            distances.append(
                self.__distance(design_point, example.design_point)
            )

        largest_list = nlargest(n, distances)
        if len(largest_list) > n - 1:
            return largest_list[n - 1]
        else:
            return largest_list[-2]

    @staticmethod
    def __mean(numbers: List[float]) -> float:
        return float(sum(numbers)) / max(len(numbers), 1)

    def __set_threshold(self, examples: List[Example]) -> None:
        distances = []
        for example in examples:
            distances.append(
                self.__nth_nearest_distance(
                    example.design_point,
                    examples,
                    int(self.__nearest_distance),
                )
            )
        print("Distances for threshold: " + str(distances))
        self.__threshold = int(round(self.__mean(distances)))
        print("Set Threshold: " + str(self.__threshold))



from pathlib import Path
from typing import List, Tuple, Optional, Dict, Iterator
from antlr4 import CommonTokenStream, FileStream
from dovado_rtl.antlr.generated.vhdl.vhdlLexer import vhdlLexer
from dovado_rtl.antlr.generated.vhdl.vhdlParser import vhdlParser
from dovado_rtl.antlr.vhdl_entity_visitor import VhdlEntityVisitor
from dovado_rtl.antlr.generated.Verilog2001.Verilog2001Parser import (
    Verilog2001Parser,
)
from dovado_rtl.antlr.generated.Verilog2001.Verilog2001Lexer import (
    Verilog2001Lexer,
)
from dovado_rtl.antlr.verilog2001_entity_visitor import (
    Verilog2001EntityVisitor,
)
from dovado_rtl.antlr.generated.SysVerilogHDL.SysVerilogHDLParser import (
    SysVerilogHDLParser,
)
from dovado_rtl.antlr.generated.SysVerilogHDL.SysVerilogHDLLexer import (
    SysVerilogHDLLexer,
)
from dovado_rtl.antlr.sysverilog_entity_visitor import (
    SysVerilogHDLEntityVisitor,
)
from dovado_rtl.antlr.hdl_representation import (
    Entity,
    PortDirection,
    PortType,
    TopLevel,
    Parameter,
    Port,
)
from dovado_rtl.abstract_classes import AbstractSourceParser
from dovado_rtl.enums import RTL
from dovado_rtl.fill_handler import FillHandler


class SourceParser(AbstractSourceParser):
    def __init__(
        self, root_folder: str, src_path: str, entity: Optional[Entity] = None
    ):
        self.__root_folder: Path = Path(root_folder)
        self.__posix_path: Path = Path(src_path)
        supported_extensions = [".vhd", ".vhdl", ".v", ".sv"]
        if self.__posix_path.suffix not in supported_extensions:
            raise Exception(
                "Supported extensions are: "
                + str(supported_extensions)
                + ", input extension "
                + self.__posix_path.suffix
                + " not supported"
            )

        self.__input_stream = FileStream(str(self.__posix_path))
        self.__RTL = (
            RTL.VHDL
            if (
                self.__posix_path.suffix == ".vhd"
                or self.__posix_path.suffix == ".vhdl"
            )
            else RTL.VERILOG
            if self.__posix_path.suffix == ".v"
            else RTL.SYSTEM_VERILOG
        )
        self.__entities, self.__top_level = self.parse()
        self.__selected_entity: Optional[Entity] = entity

    def parse(self) -> Tuple[List[Entity], Optional[TopLevel]]:
        if self.__RTL is RTL.VHDL:
            lexer = vhdlLexer(self.__input_stream)
            token_stream = CommonTokenStream(lexer)
            parser = vhdlParser(token_stream)
            visitor = VhdlEntityVisitor()
            tree = parser.design_file()
        elif self.__RTL is RTL.VERILOG:
            lexer = Verilog2001Lexer(self.__input_stream)
            token_stream = CommonTokenStream(lexer)
            parser = Verilog2001Parser(token_stream)
            visitor = Verilog2001EntityVisitor()
            tree = parser.source_text()
        else:
            lexer = SysVerilogHDLLexer(self.__input_stream)
            token_stream = CommonTokenStream(lexer)
            parser = SysVerilogHDLParser(token_stream)
            visitor = SysVerilogHDLEntityVisitor()
            tree = parser.source_text()

        visitor.visit(tree)
        return visitor.entities, visitor.top_level

    @staticmethod
    def get_project_path(file_path: Path) -> Path:
        # file_path is resolved by typer, it is always an solved path (no symlinks)
        # assumption: if a dir exists in the parent directory such that it contains a RTL file
        # the top file is considered as part of a library

        parent_content = next(file_path.parents[1].iterdir())
        if parent_content.is_dir() and next(
            parent_content.iterdir()
        ).suffix in {".vhdl", ".v", ".sv", "vhd"}:
            return file_path.parents[1].absolute()
        return file_path.parents[0].absolute()

    def get_selected_entity(self) -> Entity:
        if self.__selected_entity:
            return self.__selected_entity
        raise Exception(
            "Trying to retrieve selected entity when it is not yet set"
        )

    def get_entity(self, entity: str) -> Entity:
        for e in self.__entities:
            if e.get_name() == entity:
                return e
        raise Exception(
            "Entity "
            + entity
            + " has not been found among parsed entities "
            + str(self.__entities)
        )

    def set_entity(self, entity: str) -> None:
        self.__selected_entity = self.get_entity(entity)

    def get_top_entity(self) -> Entity:
        if self.__selected_entity:
            return self.__selected_entity
        raise ValueError(
            "Trying to access selected entity when it is not yet set"
        )

    def get_path(self) -> str:
        return str(self.__posix_path)

    def get_folder(self) -> str:
        return str(self.__root_folder.name)

    def get_hdl(self) -> RTL:
        return self.__RTL

    def get_entities(self) -> List[Entity]:
        return self.__entities

    def get_parameters(self) -> List[Parameter]:
        return self.get_top_entity().get_parameters()

    def get_parameter(self, parameter: str) -> Parameter:
        for p in self.get_top_entity().get_parameters():
            if p.get_name() == parameter:
                return p
        raise Exception(
            "Parameter "
            + parameter
            + "has not been found among parsed parameters"
            + str(self.get_top_entity().get_parameters())
            + " in entity "
            + self.get_top_entity().get_name()
        )

    def get_top_level(self):
        return self.__top_level

    def get_parameter_value(self, parameter: str) -> Optional[int]:
        return self.get_parameter(parameter).get_value()

    def __set_parameter_value(self, parameter: str, value: int) -> None:
        self.get_parameter(parameter).set_value(value)

    def write_parameter_values(
        self, hdl_handler: FillHandler, values: Dict[str, int],
    ):
        # FrameHandler is too broad, should use HdlBoxHandler but in python 3.6 is not
        # possible to solve the circular dependency

        for p, v in values.items():
            self.__set_parameter_value(p, v)

        hdl_handler.replacements = [
            p.get_name()
            for p in self.get_top_entity().get_parameters()
            if p in values.keys()
        ]
        hdl_handler.set_parameters(
            [self.get_parameter(p) for p in values.keys()]
        )
        hdl_handler.fill()

    def get_ports(self) -> List[Port]:
        return self.get_top_entity().get_ports()

    def get_port(self, port: str) -> Port:
        for p in self.get_top_entity().get_ports():
            if p.get_name() == port:
                return p
        raise Exception(
            "Port "
            + port
            + "has not been found among parsed ports"
            + str(self.get_top_entity().get_ports())
            + " in entity "
            + self.get_top_entity().get_name()
        )

    def check_port(self, port: str) -> Optional[Port]:
        for p in self.get_top_entity().get_ports():
            if p.get_name() == port:
                return p
        return None

    def get_port_direction(self, port: str) -> Optional[PortDirection]:
        return self.get_port(port).get_direction()

    def get_port_type(self, port: str) -> Optional[PortType]:
        return self.get_port(port).get_type()

    def get_imports(self) -> Tuple[List[str], List[str]]:
        return (
            self.__top_level.get_libraries(),
            self.__top_level.get_use_clauses(),
        )

    def is_library_project(self) -> bool:
        if self.__RTL is not RTL.VHDL:
            return False
        return any(p.is_dir() for p in self.__root_folder.iterdir())

    def get_user_defined_libs(self) -> Iterator[str]:
        if self.__RTL is not RTL.VHDL:
            yield from ()
        for p in self.__root_folder.iterdir():
            if p.is_dir():
                yield p.name

    def get_root_folder(self) -> Path:
        return self.__root_folder
